<div role="navigation" aria-label="breadcrumbs navigation">
  {% if page.title == 'Home' %}
    <homepagetitle>SLADE Scripting Documentation</homepagetitle>
  {% else %}
    <ul class="wy-breadcrumbs">
      <li><!--<a href="{{ nav.homepage.url }}">SLADE Scripting Documentation</a>-->Home &raquo;</li>
      {% if page %}
        {% for doc in page.ancestors %}
          {% if doc.link %}
            <li><a href="{{ doc.link|e }}">{{ doc.title }}</a> &raquo;</li>
          {% else %}
            <li>{{ doc.title }} &raquo;</li>
          {% endif %}
        {% endfor %}
      {% endif %}
      <!--{% if page %}<li>{{ page.title }}</li>{% endif %}
      <li class="wy-breadcrumbs-aside">
        {%- block repo %}
        {% if page and page.edit_url %}
          <a href="{{ page.edit_url }}"
          {%- if config.repo_name|lower == 'github' %}
            class="icon icon-github"
          {%- elif config.repo_name|lower == 'bitbucket' %}
            class="icon icon-bitbucket"
          {% endif %}> Edit on {{ config.repo_name }}</a>
        {% endif %}
        {%- endblock %}
      </li>-->
    </ul>
    {% if page %}<pagetitle>{{ page.title }}</pagetitle>{% endif %}
    <!--<hr/>-->
  {% endif %}
</div>
