
/* @import url('https://fonts.googleapis.com/css?family=Fira+Sans:400,400i,500,700,700i,800|Fira+Mono:400,500,600'); */
@import url('https://fonts.googleapis.com/css?family=Nunito+Sans:400,400i,600,800');
/* @import url('https://fonts.googleapis.com/css?family=Encode+Sans:400,400i,500,600,700,700i,800'); */
/* @import url('https://fonts.googleapis.com/css?family=Encode+Sans+Semi+Condensed:400,400i,500,600,700,700i,800'); */

body, h1, h2, h3, h4, h5, h6 {
	font-family: 'Nunito Sans', 'Segoe UI', sans-serif;
}

.wm-page-content {
	font-size: 1.1em;
	max-width: 56em;
	line-height: 1.5em;
}

.wm-page-content h1, h2, h3, h4, h5, h6, article-head {
	font-weight: 600;
}

.wm-page-content li {
	margin-bottom: 0.15em;
}

article-head {
	border-top: 1px solid;
	border-bottom: 1px solid;
	text-align: center;
	font-size: 2.1em;
	font-weight: 800;
	margin-left: -15px;
	margin-right: -15px;
	margin-top: -8px;
	margin-bottom: 1em;
	padding: 0.5em 0.5em;
	display: block;
}

.h2, h2 {
	font-size: 1.8em;
	font-weight: 800;
	border-bottom: 1px solid;
	padding: 0.3em 0;
	margin: 1.2em 0 1em 0;
}

.h3, h3 {
	font-size: 1.4em;
	font-weight: 600;
}

.h4,
h4,
listhead {
	font-size: 1.2em;
	font-weight: 600;
	margin-top: 0.8em;
}

listhead {
	display: block;
	margin-bottom: -0.4em;
}

blockquote {
	font-size: 0.95em;
}

.wm-page-content A:link,
.wm-page-content A:visited,
.wm-page-content code A:link,
.wm-page-content code A:visited {
	font-weight: 600;
	transition: all .1s ease-in;
}

.wm-page-content A:hover,
.wm-page-content A:active,
.wm-page-content code A:hover,
.wm-page-content code A:active {
	transition: all .08s ease-out;
	text-decoration: none;
}



.btn-default,
.btn-default A:link,
.btn-default A:visited,
.btn-default:focus,
.btn-default:hover,
.btn-default.active {
	border: none;
}

.btn-default A:link,
.btn-default A:visited,
.btn-default:focus {
	transition: all .1s ease-in;
}

.btn-default:hover {
	transition: all .08s ease-out;
}



footer {
	display: none;
}

hr {
	border-width: 1px;
	margin: 1.7em 0;
}




type,
arg,
prop,
prop.ro,
prop.rw,
code,
function,
func,
.admonition code,
.md-typeset code {
    font-family: 'Hack', 'Fira Mono', 'Consolas', monospace !important;
    font-size: 0.9em;
	box-shadow: none;
	font-weight: 400;
}

code,
.md-typeset code,
.admonition code {
	border: none;
	padding: 0;
	margin: 0;
}

.wm-page-content type a:link,
.wm-page-content code a:link,
.wm-page-content func a:link {
	font-weight: 400 !important;
	/* text-decoration: underline; */
}

prop.ro::before {
	content: "\25CB";
	color: hsl(10, 88%, 53%);
	margin-right: 0.4em;
	font-weight: 600 !important;
	/* font-size: 1.1em; */
}

prop.rw::before {
	content: "\25CF";
	color: hsl(140, 68%, 33%);
	margin-right: 0.4em;
	font-weight: 600 !important;
	/* font-size: 1.1em; */
}

pre code,
.codehilite pre {
	font-family: 'Hack', 'Fira Mono', 'Consolas', monospace !important;
	background: none repeat scroll 0 0 rgb(37, 37, 37) !important;
	border: 1px solid;
	line-height: 1.5em !important;
	font-size: 0.85em;
}

div .codehilite {
	margin-bottom: 0.5em;
}

pre, .rst-content tt {
	max-width: 100%;
	border: solid 1px;
	overflow-x: auto;
	font-size: 0.9em;
	line-height: 1.5em !important;
	margin: 0 !important;
}

fdef {
	font-family: 'Hack', 'Fira Mono', 'Consolas', monospace !important;
	display: block;
}

fdef > func,
fdef > arg,
fdef > type {
	font-size: 1.0em;
}

fdef > block {
	display: inline-block;
	vertical-align: text-top;
}




table {
	border: none;
	margin-top: 1em;
	width: 100%;
}

td, th {
	padding: 0.3em 0.4em;
}

.table-bordered {
	border: 0px;
}

thead {
	border-bottom: 2px solid;
}

.table-bordered > tbody > tr > td,
.table-bordered > tbody > tr > th,
.table-bordered > tfoot > tr > td,
.table-bordered > tfoot > tr > th,
.table-bordered > thead > tr > td,
.table-bordered > thead > tr > th {
	border: none;
}

.table-condensed > tbody > tr > td,
.table-condensed > tbody > tr > th,
.table-condensed > tfoot > tr > td,
.table-condensed > tfoot > tr > th,
.table-condensed > thead > tr > td,
.table-condensed > thead > tr > th {
	padding: 0.5em 0.6em;
}

.table-condensed > thead > tr > th {
	padding-bottom: 0.4em;
	padding-top: 0.5em;
}



.admonition p {
	margin: 10px 0 0 0;
}

.admonition p.admonition-title {
	margin: 0;
	font-size: 1.3em;
}





/* ToC Pane */

.wm-toc-pane {
	position: absolute;
	top: 0px;
	padding-top: 70px;
	height: 100%;
	width: 250px;
	z-index: 1;
	border-right: 0;
	overflow: auto;
	margin-left: 0px;
	padding-left: 1rem;
	padding-right: 1rem;
	padding-bottom: 2rem;
	transition: margin-left 0.3s;
	box-shadow: 0 0 10px rgba(0, 0, 0, 0.7);
}

.wm-toctree a, .wm-toctree a:visited, .wm-toctree a:hover, .wm-toctree a:focus {
	text-decoration: none;
	outline: none;
}

.wm-toc-li:hover {
	border-radius: 2px;
}

.wm-toc-li.wm-current, .wm-toc-li.wm-current:hover {
	border-radius: 2px;
}

.wm-toc-lev1,
.wm-toc-lev1 > .wm-toc-text {
	font-weight: 600;
	font-size: 1.1em;
}

.wm-toc-lev2 {
	font-weight: 500;
}
