-- An archive script will run the Execute function (below) on the current archive
-- Archive scripts can be selected and run from the "Archive->Scripts" menu

function Execute(archive)
   -- Write your archive script here
end
