
@import url('https://fonts.googleapis.com/css?family=Fira+Sans:400,400i,500,700,700i,800');
@import url('https://fonts.googleapis.com/css?family=Fira+Mono:500,600');

li,
p {
	font-size: 0.97em;
	line-height: 1.5em;
}

body, h1, h2, h3, h4, h5, h6 {
	font-family: 'Fira Sans', 'Segoe UI', sans-serif;
}

.wy-nav-content {
	background-color: hsl(220, 100%, 99%);
	box-shadow: 0px 0px 20px hsl(220, 35%, 71%);
}

.wy-nav-content-wrap {
	background-color: hsl(220, 100%, 94%);
}

@media screen and (min-width: 1400px) {
	.wy-nav-content {
		margin-left: auto;
		margin-right: auto;
	}
}

a {
    font-weight: 500;
}

.wy-nav-content A:link,
.wy-nav-content A:visited,
.wy-nav-content code A:link,
.wy-nav-content code A:visited {
	color: hsl(220, 90%, 40%);
}

.wy-nav-content A:hover,
.wy-nav-content A:active,
.wy-nav-content code A:hover,
.wy-nav-content code A:active {
	color: hsl(40, 100%, 45%);
}

h1,
h2,
h3,
h4,
h5 {
	color: hsl(220, 28%, 43%);
}

h2,
h1 {
	border-bottom: 0.1em solid hsl(220, 35%, 90%);
    padding-bottom: 0.2em;
}

hr {
	border-top: 0.1em solid hsl(220, 35%, 90%);
}

type,
arg,
prop,
code,
function,
.admonition code {
    font-family: 'Fira Mono', 'Consolas', monospace;
    font-size: 0.9em;
}

code,
.admonition code {
	color: hsl(220, 28%, 43%);
	border: none;
	background-color: inherit;
	padding: 0;
}

div.codehilite {
	font-size: 0.8em;
	overflow: auto;
	background-color: hsl(220, 58%, 98%);
	border: 1px solid hsl(220, 58%, 90%);
	margin-bottom: 1.5em;
	padding: 0 0.5em;
}

type,
type a {
	/*color: hsl(40, 98%, 33%) !important;*/
	color: hsl(30, 98%, 33%) !important;
}

type a {
	font-weight: 500;
	text-decoration: underline;
}

type a:hover {
	/*color: hsl(40, 100%, 45%) !important;*/
	color: hsl(30, 100%, 45%) !important;
}

arg,
prop {
	font-weight: 600;
	color: hsl(270, 68%, 43%);
}

prop {
	color: hsl(170, 68%, 33%);
}

pre code {
	background: none repeat scroll 0 0 hsl(220, 93%, 97%) !important;
	border: 1px solid hsl(220, 12%, 88%);
	line-height: 1.5em !important;
}

.hljs-keyword {
	color: hsl(270, 68%, 43%);
}

.hljs-string {
	color: hsl(345, 76%, 40%);
}

.hljs-comment {
	color: hsl(100, 30%, 50%);
}

.hljs-built_in {
	color: hsl(200, 78%, 43%);
}

.wy-nav-top,
.wy-side-nav-search {
	background-color: hsl(220, 27%, 57%);
	box-shadow: 0 0 20px rgba(0, 0, 0, 0.5);
}

a.btn.btn-neutral {
	border: none;
	background-color: hsl(220, 27%, 57%) !important;
	color: white !important;
	box-shadow: none;
}

a.btn.btn-neutral:hover {
	background-color: hsl(220, 27%, 47%) !important;
}

table.docutils td,
table.docutils th {
	border: none !important;
}

table.docutils thead {
	color: hsl(220, 57%, 97%) !important;
	background-color: hsl(220, 27%, 57%) !important;
	font-size: 1.1em;
}

table.docutils tbody td {
	background-color: inherit !important;
}

table.docutils tbody tr:nth-child(odd) {
	background-color: hsl(220, 60%, 96%) !important;
}

blockquote {
	border-left: 0.3em solid hsl(220, 27%, 57%);
	color: hsl(220, 27%, 55%);
	padding-left: 1em;
	margin-left: 1em !important;
}

blockquote p {
	margin: 0;
}

indent {
	margin-left: 2em;
}

function {
	display: block;
	margin-bottom: -1.2em !important;
	background-color: hsl(220, 27%, 57%) !important;
	color: white !important;
	padding: 0.2em 0.5em;
}

args {
	display: block;
	background-color: hsl(220, 60%, 95%) !important;
	padding: 0.2em 0;
	border: 1px solid hsl(220, 27%, 77%);
}

args type {
	padding-left: 2em;
}

params,
listhead {
	font-weight: 700;
	display: block;
	margin-bottom: -1.3em;
}

pagetitle,
homepagetitle {
	font-family: 'Fira Sans', 'Segoe UI', sans-serif;
	font-size: 2em;
	font-weight: 900;
	padding: 0em 0.5em 0.2em 0.5em;
	display: block;
	background-color: hsl(220, 27%, 57%) !important;
	color: white !important;
	margin-bottom: 0.5em;
	margin-top: 0em;
	border-radius: 3px;
	border-top-left-radius: 0;
	border-top-right-radius: 0;
}

homepagetitle {
	border-top-left-radius: 3px;
	border-top-right-radius: 3px;
	padding: 0.3em 0.5em;
}

ul.wy-breadcrumbs {
	font-weight: 500;
	background-color: hsl(220, 27%, 57%) !important;
	color: hsl(220, 77%, 93%) !important;
	padding: 0.5em 1em 0.2em 1em;
	border-radius: 3px;
	border-bottom-left-radius: 0;
	border-bottom-right-radius: 0;
}

.wy-side-nav-search input[type=text] {
	border-radius: 8px;
	border: none;
}

li.current .subnav,
.subnav .subnav {
	background-color: hsl(0, 3%, 17%);
}

li.current .subnav li a:hover,
.subnav .subnav li a:hover {
	background-color: hsl(0, 3%, 19%);
}

.wy-menu-vertical li.current>a {
	background-color: hsl(220, 28%, 43%);
	color: hsl(222, 100%, 96%);
}

.wy-menu-vertical li.current>a:hover {
	background-color: hsl(220, 28%, 40%);
	color: hsl(222, 100%, 96%);
}

.rst-content .warning,
.rst-content .attention,
.rst-content .note,
.rst-content .tip {
	border-radius: 3px;
}

.rst-content .warning .admonition-title,
.rst-content .attention .admonition-title,
.rst-content .note .admonition-title,
.rst-content .tip .admonition-title {
	border-top-left-radius: 3px;
	border-top-right-radius: 3px;
}

/*
function::after {
	font-size: 0.8em;
	font-weight: 400;
	content: " (";
}

args::after {
	content: ")";
}

/*
.wy-side-nav-search a {
    color: hsl(220, 100%, 94%);
}

.wy-nav-side {
    background-color: hsl(220, 100%, 94%);
    color: black;
}

.wy-menu-vertical a {
    color: black;
}

.wy-menu-vertical a:hover {
    background-color: hsl(220, 47%, 87%);
}

span.caption-text {
    color: hsl(220, 47%, 40%);
    background-color: hsl(220, 97%, 92%);
    border-bottom: 1px solid hsl(220, 97%, 90%);
}

.wy-menu-vertical .subnav a,
.wy-menu-vertical .subnav .subnav a {
    color: black;
}

.wy-menu-vertical .subnav .subnav a:hover {
    background-color: hsl(220, 77%, 93%);
}

li.current a,
.subnav .subnav {
    background-color: hsl(220, 77%, 97%);
    color: black !important;
}

li.current a:hover {
    background-color: hsl(220, 77%, 93%) !important;
}

.rst-versions {
    background-color: hsl(220, 27%, 37%);
}

.rst-versions .rst-current-version {
    background-color: hsl(220, 27%, 57%);
}

.rst-versions .rst-current-version a:hover {
    color: white;
}*/
