-- An entry script will run the Execute function (below) on the current/selected entries
-- Entry scripts can be selected and run from the "Entry->Scripts" menu,
-- or "Run Script" in the entry list context menu

function Execute(entries)
   for i,entry in ipairs(entries) do
      -- Write your entry script here
   end
end
