#pragma once

// Forward declare a bunch of basic, often-used wx controls
class wxBitmapButton;
class wxButton;
class wxCheckBox;
class wxChoice;
class wxComboBox;
class wxMenu;
class wxRadioButton;
class wxSizer;
class wxSlider;
class wxSpinCtrl;
class wxStaticBox;
class wxStaticText;
class wxTextCtrl;
