-- A map script will run the Execute function (below) on the current map
-- Map scripts can be selected and run from the "Tools->Run Script" menu in the map editor

function Execute(map)
   -- Write your map script here
end
